/* -*- linux-c -*- */


/*
    lm78.c - A Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
 *  LM78 sensor driver for LINUX
 *
 * Copyright (c) 1997 Alexander Larsson
 * 
 * provides a /proc/sensors which shows
 * some information about your system.
 * Voltages, mb temperature and fan speeds
 * are currently reported.
 *
 * This should work on all ASUS TX97 cards
 * but it's only tested on a TX97-X.
 * It should also work on any card that has
 * a LM78 circuit, but some constants in
 * lm78.h probably has to be changed.
 *
 * please report any bugs/patches to me:
 *  alla@lysator.liu.se
 *
 * Heavily patched by Frodo Looijaard <frodol@dds.nl> and Philip 
 * Edelbrock <phil@Stimpy.netroedge.com>.
 *
 * Reference:
 *   National Semiconductor
 *   LM78 Microprocessor System Hardware Monitor, September 1996.
 *   http://www.national.com/ds/LM/LM78.pdf
 *
 */

#define LM78_VERSION  "0.01"

#include <linux/config.h>

/* Added 19980620 by Frodo Looijaard (frodol@dds.nl)
   This solves undefined symbol errors for depmod/modprobe.
   Tested for Linux 2.0.34, but I do not know whether it works for other
   kernel versions. Actually, I am not even sure this is the best way to
   solve this. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/bios32.h>
#include <asm/io.h>
#include <linux/pci.h>
#include "lm78.h"
#include "lm75.h"

static int Init_LM75(void);
static void Cleanup_LM75(void);
static int read_smbus_word(int addr);

/* work around for now */
static void our_usleep(int wait_time);

unsigned short PIIX4_smba = 0;
int LM75_Disabled = 0;

int init_module(void);
void cleanup_module(void);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
static int lm78_read(char *buf, char **start, off_t offset,
                     int len, int *eof, void *private);
#else
static int lm78_read(char *, char **, off_t, int, int);
#endif
static int lm78_read_value(int address);

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
MODULE_AUTHOR("Alexander Larsson <alla@lysator.liu.se>, Frodo Looijaard <frodol@dds.nl>, Philip Edelbrock <phil@Ren.netroedge.com>");
MODULE_DESCRIPTION("Driver for reading sensor values from a lm78 on the motherboard.");
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
 static struct proc_dir_entry *proc_sensors = NULL;
#else
 static int inodeval = 0;
 struct proc_dir_entry dir =
 {
   0, 7, "sensors", 
   S_IFREG | S_IRUGO, 1, 0, 0,
   0, NULL,
   &lm78_read
 };
#endif

 static struct sensor_data sensor_data = {
   0
 };


/* This kill some time in a baaaaaad way. :') */
static void our_usleep(int wait_time) {
long killtime;
long temp;
 for (; wait_time > 0; wait_time--) {
	 for (killtime=0; killtime<=40000; killtime++) {
		temp=0xAABC; temp= (temp << 13);
         }
  }
}




/*
 *	Initialize LM75/SMBus code.
 */
int Init_LM75(void){
int error_return=0;
unsigned char  PIIX4_bus, PIIX4_devfn;

    if (pcibios_present() == 0) {
        printk("LM78: Error: No PCI-bus found!\n");
	error_return=1;
    } else {    
        if(pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB_0, 0,
                               &PIIX4_bus, &PIIX4_devfn)) {
            printk("LM78: Error: Cant detect PIIX4!\n");
	    error_return=1;
        } else {
            pcibios_read_config_word(PIIX4_bus, PIIX4_devfn | 3, 0x90, &PIIX4_smba);
            PIIX4_smba &= 0xfff0;   
	    if (check_region(PIIX4_smba, 7))
		return 0; 
            request_region(PIIX4_smba, 7, "lm78");
        }   
    }   
    if (error_return)
	printk("LM78: SMBus setup failed.  LM75 readings will be disabled.\n");
/* This line is for debugging */
    printk("PIIX4_smba = 0x%X\n",PIIX4_smba);
    return error_return;
}

/*
 *      Find and read LM75 at given address.
 */
int read_smbus_word(int addr) {

int HOST_BUSY;
long timeout=0;
int result=0;

/* Setup protocol type, address, device command, data, and control bits */
    outb_p( 0x00, SMBHSTDAT0);
    outb_p( 0x00, SMBHSTDAT1);
    outb_p( 0x0C, SMBHSTCNT); /* Read word */
    if (inb_p(SMBHSTSTS) != 0x00) {
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
    }
    outb_p( (addr << 1) | 0x01, SMBHSTADD); /* address of possible thermal sensors are from 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); /* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   our_usleep(100);

/* Pole Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (((inb_p(SMBHSTSTS) & 0x01)) & (timeout < MAX_TIMEOUT)) {
/* Wait for a while */
        our_usleep(50);
   }
   if (! (inb(SMBHSTSTS) & 0x01)) {
        result=(inb(SMBHSTDAT0) << 8) | inb(SMBHSTDAT1);
   }

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
   }
   return result;
}

/*
 *      Clean-up the LM75/SMBus code.
 */
void Cleanup_LM75(void){
        release_region(PIIX4_smba, 7); 
}

/*
 *	Initialize the module.
 */
int init_module(void)
{
  int res;
  
  if (check_region(LM78_BASE, LM78_EXTENT)) 
    return -ENODEV;

  /* Verify that lm78 exists: */
  outb_p( LM78_DUMMY, LM78_ADDRESS_REG );
  res = inb_p(LM78_ADDRESS_REG);
  inb_p(LM78_DATA_REG); /* reset busy flag */

  if ((res & 0x7F) != LM78_DUMMY) 
    return -EIO; 

  request_region(LM78_BASE, LM78_EXTENT, "LM78");

  /* Reset all except Watchdog values and last conversion values 
     This sets fan-divs to 2, among others */
  LM78_WRITE(LM78_CONFIG, 0x80 );

  /* Start monitoring */
  LM78_WRITE(LM78_CONFIG, (LM78_READ(LM78_CONFIG) & 0xf7) | 0x01);

  sensor_data.last_read=jiffies;
  sensor_data.valid=0;

  /* register /proc/sensors */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
  
  proc_sensors = create_proc_entry("sensors", 0, 0);
  if (proc_sensors) {
    proc_sensors->read_proc = lm78_read;
  } else {
    printk("Couldn't create /proc/sensors\n");
    release_region(LM78_BASE, LM78_EXTENT);
    return -ENOENT;
  }
  
#else
  
/*
 *	Try to register the file in the root directory of the proc
 *	filesystem.  Dynamic registration means we won't provide a fixed
 *	inode number, but rather the registration code will allocate one for
 *	us.  This is safer, since we never know when someone will add a new
 *	fixed inode number to the filesystem.  Possible returns are 0
 *	(success) and -ENOENT (failure), and since there's only one error we
 *	won't worry unduly about messages.  If successful, continue.
 */
  if (proc_register_dynamic(&proc_root, &dir)) {
    release_region(LM78_BASE, LM78_EXTENT);
    return -ENOENT;
  }
  
/*
 *      Save the inode value from the registration call, as we will need
 *      it to unregister.
 */
  inodeval = dir.low_ino;

#endif
  
  printk("LM78 detected and installed.\n");
  if ((LM75_Disabled = Init_LM75()) == 0)
	printk("SMBus/LM75 detected and installed.\n");
  return 0;
}

/*
 *	Remove the module.  If a nonzero inode value was returned,
 *	unregister the /proc directory entry; that's all we need to do.
 */

void cleanup_module(void)
{
        release_region(LM78_BASE, LM78_EXTENT);
	Cleanup_LM75();
      
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
        if (proc_sensors)
		proc_unregister(&proc_root, proc_sensors->low_ino);
#else
        if (inodeval)
		proc_unregister(&proc_root, inodeval);
#endif
}

/*
 *	Process a read to the file.  we just return the whole mess to the
 *	user with a new-line character on the end.
 */

static int lm78_read_value(int address) {
  while (LM78_BUSY) {
    current->state = TASK_INTERRUPTIBLE;
    current->timeout = jiffies + 10;
    schedule();
  }
  
  return LM78_READ(address);
}


static void read_sensor_data()
{
  /* Note that check for the jiffies counter overflow. Note that we
     assume that when it overflows, it will be reset to 0. If not,
     we may probe once too many (too bad). */
  if ((jiffies - sensor_data.last_read < 1*HZ ) && 
      !((jiffies < sensor_data.last_read) && (jiffies < 1*HZ)))
    return; /* Less than one second ago. Not updated anyway. */

  /* TODO: grab Spin lock on sensor_data */

  sensor_data.valid = 1; 

#ifdef LM78_VCORE
  sensor_data.Vcore = ((long) lm78_read_value(LM78_VCORE)) * 16 * NORMAL_VCORE / 100000;
#endif

#ifdef LM78_3_3V
  sensor_data.V3_3v = ((long) lm78_read_value(LM78_3_3V)) * 16 * NORMAL_3_3V / 100000;
#endif

#ifdef LM78_5V
  sensor_data.V5v = ((long) lm78_read_value(LM78_5V)) * 16 * NORMAL_5V / 100000;
#endif

#ifdef LM78_M5V
  sensor_data.Vm5v = ((long) lm78_read_value(LM78_M5V)) * 16 * NORMAL_M5V / 100000;
#endif

#ifdef LM78_12V
  sensor_data.V12v = ((long) lm78_read_value(LM78_12V)) * 16 * NORMAL_12V / 100000;
#endif

#ifdef LM78_M12V
  sensor_data.Vm12v = ((long) lm78_read_value(LM78_M12V)) * 16 * NORMAL_M12V / 100000;
#endif
  
  sensor_data.temp_mb= lm78_read_value(LM78_TEMPERATURE);

  sensor_data.fan[0] = LM78_READ( LM78_FAN1 );
  sensor_data.fan[1] = LM78_READ( LM78_FAN2 );
  sensor_data.fan[2] = LM78_READ( LM78_FAN3 );

  sensor_data.last_read = jiffies;
  
  /* TODO: release Spin lock on sensor_data */
}


/*
 * Ahem, gotta find out how this is supposed to work, ie. how big is 'buf' 
 */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
static int lm78_read(char *buf, char **start, off_t offset, int len, int *eof, void *private)
#else
static int lm78_read(char *buf, char **start, off_t offset, int len, int unused)
#endif
{
  int fan,i,value;
  long v;
  len=0;

  read_sensor_data();

  if (sensor_data.valid) {
  
#ifdef LM78_VCORE
    v = sensor_data.Vcore;
    len += sprintf(buf+len, "Vcore:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "Vcore: not supported\n");
#endif

#ifdef LM78_3_3V
    v = sensor_data.V3_3v;
    len += sprintf(buf+len, "+3.3V:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+3.3V: not supported\n");
#endif

#ifdef LM78_5V
    v = sensor_data.V5v;
    len += sprintf(buf+len, "+5.0V:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+5.0V: not supported\n" );
#endif

#ifdef LM78_M5V
    v = sensor_data.Vm5v;
    len += sprintf(buf+len, "-5.0V: %3d.%02dV\n", -(int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "-5.0V: not supported\n" );
#endif

#ifdef LM78_12V
    v = sensor_data.V12v;
    len += sprintf(buf+len, "+12V:   %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+12V : not supported\n" );
#endif

#ifdef LM78_M12V
    v = sensor_data.Vm12v;
    len += sprintf(buf+len, "-12V:  %3d.%02dV\n", (int) -v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "-12V : not supported\n" );
#endif
  
    len += sprintf(buf+len, "Mainboard temperature: %d degrees\n",
                   sensor_data.temp_mb );

    fan = sensor_data.fan[0];
    len += sprintf(buf+len, "Chassis fan: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


    fan = sensor_data.fan[1];
    len += sprintf(buf+len, "CPU fan: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


    fan = sensor_data.fan[2];
    len += sprintf(buf+len, "Power fan: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));

    /* Tack on LM75 stats */
    if (! LM75_Disabled) {
    /* Probe each possible valid LM75 address and report the temp for each found */
	for (i=0x048; i<=0x04F; i++) { 
        value=read_smbus_word(i) >> 7;
        if (value != -1) {
	   len += sprintf(buf+len, "LM75 at address 0x%X reports %d.%d degrees C\n", i, value >> 1, (value & 1) * 5);
        }
     }
    }
  }
  else
    len += sprintf(buf+len, "No sensor data yet (try again in a few moments)\n");

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
  *start = buf + offset;
  return len > offset ? len - offset : 0;
#else
  return len;
#endif
}

